class LimitedIntSet{

      var limit : int;
      var arr : array<int>;
      var size : int;
      
      predicate Valid()
      reads this, this.arr;
      {arr != null && 0 <= size <= limit && limit == arr.Length}
      
      predicate Sorted()
      reads this, this.arr;
      requires Valid();
      {forall i, j :: 0 <= i < j < size ==> arr[i] <= arr[j]}

      method Init(l : int)
        modifies this;
        requires l > 0;
        ensures fresh(arr);
        ensures Valid();
        ensures limit == l && size == 0;
     {   
      limit := l;
      arr := new int[l];
      size := 0;   
     }
     
     function method contains (elem : int) : bool
     reads this, this.arr;
     requires this.Valid();
     {exists i :: 0 <= i < size && arr[i] == elem}
     
    
     method Add(elem : int) returns (res :bool)
     modifies this.arr, this`size;
     requires this.Valid();
     ensures Valid();
     ensures (!old(this.contains(elem)) && old(size) < limit) ==> 
             res && contains(elem) && size == old(size)+1
             && (forall e :: e!=elem && old(contains(e)) ==> contains(e)); 
     ensures (old(this.contains(elem)) || old(size) >= limit) ==> 
             !res && size == old(size) && (forall e :: old(this.contains(e)) <==> contains(e));
     {
       if (!this.contains(elem) && size < limit)
       { 
         arr[size] := elem;
         size := size +1; 
         res := true;
       }
       else
        {
          res := false;
        }
     }
  
    method Find(elem : int) returns (index : int)
    requires Valid();
    ensures 0 <= index ==> index < size && arr[index] == elem;
    ensures index < 0 ==> forall k :: 0 <= k < size ==> arr[k] != elem;  
    {
        index := 0;
        while (index < size)
        invariant 0 <= index <= size;
        invariant forall k :: 0 <= k < index ==>arr[k] != elem;
        {
          if(arr[index] == elem) {return;}
          index := index + 1;
        }
        index := -1;
    }

  method BinarySearch(elem : int) returns (index : int)
  requires Valid() && Sorted();
  ensures 0 <= index ==> index < size && arr[index] == elem;
  ensures index < 0 ==> forall k :: 0 <= k < size ==> arr[k] != elem;  
  {
    var low := 0;
    var high := size;
    while(low < high)
    invariant 0<=low <= high <= size;
    invariant forall i :: 0 <= i < size && !(low <= i < high) ==> arr[i] != elem;
    {
      var mid := (low + high) / 2;
      if(arr[mid] < elem){
        low := mid + 1;
      } else if (elem < arr[mid]){
        high := mid;
      } else{
        index := mid;
        return;
      }
    }
    index := -1;
  }
        

}


